#!/bin/bash
# ffoo - Fast Foo - a Bash Library
# See LICENSE file for copyright and license details.

tmp=$(mktemp)
sed -e 's/ = /=/' < config.mk > $tmp
. $tmp
rm -f $tmp

bindir=${DESTDIR}${PREFIX}/bin
mandir=${DESTDIR}${MANPREFIX}
docdir=${DESTDIR}${PREFIX}/share/doc/ffoo
shrdir=${DESTDIR}${PREFIX}/share/ffoo

list_of_bins() {
    echo bin/ffdoc
    echo bin/fff
    echo bin/ffoom
}

list_of_installed_bins() {
    list_of_bins | sed -e "s/bin/$(sed -e 's/\//\\\//g' <<<$bindir)/"
}

list_of_docs() {
    echo doc/fff.1.md
    echo doc/ffoo.3.md
}

list_of_shrs() {
    echo ffoo.sh
    echo include
}

build1() {
    local srcpath dstpath
    srcpath=$1
    dstpath=${srcpath%.skel}
    case $dstpath in
        *.md)   cat $srcpath \
                  | expand_includes \
                  | expand_variables \
                > $dstpath ;;
        *)      cat $srcpath \
                  | expand_variables \
                > $dstpath ;;
    esac
    echo $dstpath >> built.list
}

build() {
    local srcpath
    find -type f -name '*.skel' \
     | while read srcpath;
       do
           build1 "$srcpath"
       done
}

clean() {
    test -f built.list && {
        cat built.list | xargs rm -f
        rm -f built.list
    } || :
}

dist() {
    local version=$(get_version)
    local dirname=ffoo-$version
    mkdir -p $dirname
    cp -R   bin \
            config.mk \
            doc \
            include \
            Makefile \
            ffoo.sh.skel \
            README \
            LICENSE \
            setup \
            test \
            $dirname
    sed -ie "s/^VERSION = .*/VERSION = $version/" $dirname/config.mk
    tar -cf $dirname.tar $dirname
    gzip $dirname.tar
    rm -rf $dirname
}

expand_includes() {
    #
    # Expand include directives
    #
    # Expand e.g. `<!-- include4: foo.sh -->` to include code of foo.sh
    #
    perl -we '
        use strict;
        my $text;
        while (<>) {
            chomp;
            if (m/<!-- include4: (\S+) -->/) {
                open my $fh, $1 or warn "cannot find: $1";
                my $text = do { local($/); <$fh> };
                close $fh;
                $text =~ s/^(.)/    $1/gm;
                chomp $text;
                print "$text\n";
            } else {
                print "$_\n";
            }
        }
    '
}

expand_variables() {
    #
    # Expand variable values
    #
    perl -pe "
        s|__FFOO_DIR__|$shrdir|;
        s|__CODENAME__|$CODENAME|;
        s|__VERSION__|$(get_version)|;
    "
}

install() {
    mkdir -vp $bindir \
              $docdir \
              $shrdir
    list_of_bins | xargs cp -vrt $bindir
    list_of_docs | xargs cp -vrt $docdir
    list_of_shrs | xargs cp -vrt $shrdir
    list_of_installed_bins | xargs chmod -v 755
    test -f .autoclean && clean || :
}

manpages() {
    ronn -r doc/*.[1-9].md
    ls doc/*.[1-9] >> built.list
}

install_manpages() {
    cp -vr doc/*.1 $mandir/man1
    cp -vr doc/*.3 $mandir/man3
    test -f .autoclean && clean || :
}

get_version() {
    #
    # Build semver version string with build metadata
    #
    # Build version string from available info using following
    # logic:
    #
    #  1. use VERSION (from config.mk)
    #  2. if we are in git, override the version with last tag
    #  3. if set, add STAGE (from config.mk) as pre-release ID
    #     (afer dash)
    #  4. if we are at a later commit than the last tag, add commit
    #     sha1 to build metadata (after plus sign)
    #  5. if the tree is "dirty", i.e. has uncommited changes,
    #     add "dirty" to build metadata
    #
    # The version is compatible with SemVer 2.0.0.
    #
    # Examples:
    #
    #     myprog v1.0.7                         # all clear
    #     myprog v1.0.7-alpha                   # STAGE="alpha"
    #     myprog v1.0.7-alpha+g1aef811          # ^^ + some commits after
    #     myprog v1.0.7-alpha+g1aef811.dirty    # ^^ + tree edited
    #     myprog v1.0.7-alpha+dirty             # tag OK but tree edited
    #     myprog v1.0.7+dirty                   # ^^ but no stage
    #
    # Note that versions with "dirty" should be perceived as kind of
    # dangerous outside developer's own machine.  Versions with sha1 are
    # safer but must not be released.
    #
    # I have considered decorating the git commit refs to make them
    # sort of sortable (e.g. "r1.g1aef811"), but on second thought,
    # I don't think it's good idea to give *any* semantics to meta-data
    # at all.  First, there is no rule that r1<r2<r3; a commit can be
    # removing what other just added and one change can be split to
    # multiple commits.  Also, the whole thing breaks anyway once you
    # rebase your branch (no, it's not a sin).  The sole purpose of
    # meta-data is to *identify* the code, and provide safe path back
    # to tree; commit refs are already perfect for that.
    #
    local version=$VERSION
    local stage=$STAGE
    if git rev-parse HEAD >&/dev/null;
    then    # we are in git repo... so we can get smart
        local lasttag=$(git tag | grep ^v | sort -V | tail -n1)
        if ! git describe --tags --exact-match HEAD >&/dev/null;
        then    # we are not at later commit than the last tag
            local sha=g$(git log -1 --pretty=format:%h HEAD)
        fi
        if test "$(git diff --shortstat 2>/dev/null)" != "";
        then    # thr tree is "dirty", i.e. has been edited
            local dirty=dirty
        fi
        version=${lasttag:1}
        local suffix=""
        case $stage:$sha:$dirty in
            ::)         suffix=""                    ;;
            ::dirty)    suffix="+$dirty"             ;;
            :g*:)       suffix="+$sha"               ;;
            :g*:dirty)  suffix="+$sha.dirty"         ;;
            *::)        suffix="-$stage"             ;;
            *::dirty)   suffix="-$stage+$dirty"      ;;
            *:g*:)      suffix="-$stage+$sha"        ;;
            *:g*:dirty) suffix="-$stage+$sha.$dirty" ;;
        esac
        version=$version$suffix
    fi
    echo $version
}

uninstall() {
    list_of_installed_bins | xargs rm -vf
    rm -vrf $shrdir
    rm -vrf $docdir
    rm -vf $mandir/man1/fff.1
    rm -vf $mandir/man3/ffoo.3
}

case $1 in
    build|clean|dist|install|install_manpages|manpages|uninstall)
        $1
        ;;
    *)
        echo "usage: $(basename $0) build|clean|dist|install|uninstall" >&2
esac
